
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.23.0rc2',
    'branch_nick': 'feature-0-Release__1_23_0rc2_preparation',
    'revision_id': 'd67e601da1bd6c0eaf94a9aa40e51624172df83d',
    'revno': 9921
}
